from functools import partial
from typing import Optional, Tuple

import numpy as np

from gdsfactory.component import Component
from gdsfactory.config import logger
from gdsfactory.port import Port, read_port_markers, sort_ports_clockwise
from gdsfactory.snap import snap_to_grid
from gdsfactory.types import Layer


def add_ports_from_markers_square(
    component: Component,
    pin_layer: Layer = (69, 0),
    port_layer: Optional[Layer] = None,
    orientation: Optional[int] = 90,
    min_pin_area_um2: float = 0,
    max_pin_area_um2: float = 150 * 150,
    pin_extra_width: float = 0.0,
    port_names: Optional[Tuple[str, ...]] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
) -> Component:
    """Add ports from square markers at the port center in port_layer

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: for port markers.
        port_layer: for the new created port.
        orientation: in degrees 90: north, 0: east, 180: west, 270: south
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2
        max_pin_area_um2: ignore pins for area above certain size
        pin_extra_width: 2*offset from pin to straight
        port_names: names of the ports (defaults to {i})
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical
        port_type: optical, electrical

    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    port_markers = read_port_markers(component, [pin_layer])
    port_names = port_names or [
        f"{port_name_prefix}{i+1}" for i in range(len(port_markers.polygons))
    ]
    layer = port_layer or pin_layer

    for port_name, p in zip(port_names, port_markers.polygons):
        dy = snap_to_grid(p.ymax - p.ymin)
        dx = snap_to_grid(p.xmax - p.xmin)
        x = p.x
        y = p.y
        if dx == dy and max_pin_area_um2 > dx * dy > min_pin_area_um2:
            component.add_port(
                port_name,
                midpoint=(x, y),
                width=dx - pin_extra_width,
                orientation=orientation,
                layer=layer,
            )
    return component


def add_ports_from_markers_center(
    component: Component,
    pin_layer: Layer = (1, 10),
    port_layer: Optional[Layer] = None,
    inside: bool = False,
    tol: float = 0.1,
    pin_extra_width: float = 0.0,
    min_pin_area_um2: Optional[float] = None,
    max_pin_area_um2: float = 150.0 * 150.0,
    skip_square_ports: bool = False,
    xcenter: Optional[float] = None,
    ycenter: Optional[float] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
    auto_rename_ports: bool = True,
) -> Component:
    """Add ports from rectangular pin markers.

    markers at port center, so half of the marker goes inside and half ouside the port.

    guess port orientation from the component center (xcenter)

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: GDS layer for maker [int, int].
        port_layer: for the new created port
        inside: True-> markers  inside. False-> markers at center
        tol: tolerance for comparing how rectangular is the pin
        pin_extra_width: 2*offset from pin to straight
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2
        max_pin_area_um2: ignore pins for area above certain size
        skip_square_ports: skips square ports (hard to guess orientation)
        xcenter: for guessing orientation of rectangular ports
        ycenter: for guessing orientation of rectangular ports
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical ports.
        port_type: type of port (optical, electrical ...)
        auto_rename_ports:

    For inside=False the port location is at the middle of the PIN

    .. code::
           _______________
          |               |
          |               |
         |||             |||____  | pin_extra_width/2 > 0
         |||             |||
         |||             |||____
         |||             |||
          |      __       |
          |_____|__|______|
                |__|


    For inside=True all the pin is inside the port

    .. code::
           _______________
          |               |
          |               |
          |_              |
          | |             |
          |_|             |
          |               |
          |      __       |
          |_____|__|______|



    dx < dy: port is east or west
        x > xc: east
        x < xc: west

    dx > dy: port is north or south
        y > yc: north
        y < yc: south

    dx = dy
        x > xc: east
        x < xc: west

    """
    xc = xcenter or component.x
    yc = ycenter or component.y
    xmax = component.xmax
    xmin = component.xmin
    ymax = component.ymax
    ymin = component.ymin

    port_markers = read_port_markers(component, layers=(pin_layer,))
    layer = port_layer or pin_layer
    port_locations = []

    ports = {}

    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default

    for i, p in enumerate(port_markers.polygons):
        port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i
        dy = p.ymax - p.ymin
        dx = p.xmax - p.xmin
        x = p.x
        y = p.y

        if min_pin_area_um2 and dx * dy < min_pin_area_um2:
            logger.debug(f"skipping port at ({x}, {y}) with min_pin_area_um2 {dx * dy}")
            continue

        if max_pin_area_um2 and dx * dy > max_pin_area_um2:
            continue

        if skip_square_ports and snap_to_grid(dx) == snap_to_grid(dy):
            logger.debug(f"skipping square port at ({x}, {y})")
            continue

        pxmax = p.xmax
        pxmin = p.xmin
        pymax = p.ymax
        pymin = p.ymin

        if dx < dy and x > xc:  # east
            orientation = 0
            width = dy
            x = p.xmax if inside else p.x
        elif dx < dy and x < xc:  # west
            orientation = 180
            width = dy
            x = p.xmin if inside else p.x
        elif dx > dy and y > yc:  # north
            orientation = 90
            width = dx
            y = p.ymax if inside else p.y
        elif dx > dy and y < yc:  # south
            orientation = 270
            width = dx
            y = p.ymin if inside else p.y

        # square port markers have same width and height
        # check which edge (E, W, N, S) they are closer to

        elif pxmax > xmax - tol:  # east
            orientation = 0
            width = dy
            x = p.xmax if inside else p.x
        elif pxmin < xmin + tol:  # west
            orientation = 180
            width = dy
            x = p.xmin if inside else p.x
        elif pymax > ymax - tol:  # north
            orientation = 90
            width = dx
            y = p.ymax if inside else p.y
        elif pymin < ymin + tol:  # south
            orientation = 270
            width = dx
            y = p.ymin if inside else p.y

        elif pxmax > xc:
            orientation = 0
            width = dy
            x = p.xmax if inside else p.x

        elif pxmax < xc:
            orientation = 180
            width = dy
            x = p.xmin if inside else p.x

        x = snap_to_grid(x)
        y = snap_to_grid(y)
        width = np.round(width - pin_extra_width, 3)

        if (x, y) not in port_locations:
            port_locations.append((x, y))
            ports[port_name] = Port(
                name=port_name,
                midpoint=(x, y),
                width=width,
                orientation=orientation,
                layer=layer,
                port_type=port_type,
            )

    ports = sort_ports_clockwise(ports)

    for port_name, port in ports.items():
        if port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )

        else:
            component.add_port(name=port_name, port=port)
    if auto_rename_ports:
        component.auto_rename_ports()
    return component


add_ports_from_markers_inside = partial(add_ports_from_markers_center, inside=True)


def add_ports_from_labels(
    component: Component,
    port_width: float,
    port_layer: Layer,
    xcenter: Optional[float] = None,
    port_name_prefix: Optional[str] = None,
    port_type: str = "optical",
    get_name_from_label: bool = False,
    layer_label: Optional[Layer] = None,
    fail_on_duplicates: bool = False,
) -> Component:
    """Add ports from labels.
    Assumes that all ports have a label at the port center.
    because labels do not have width, you have to manually specify the ports width

    Args:
        component: to read polygons from and to write ports to.
        port_width: for ports.
        port_layer: for the new created port.
        xcenter: center of the component, for guessing port orientation.
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical
        port_type: optical, electrical
        layer_label:
        fail_on_duplicates: raises ValueError for duplicated port names.
            if False adds incremental suffix (1, 2 ...) to port name.

    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    xc = xcenter or component.x
    yc = component.y

    port_name_to_index = {}

    for i, label in enumerate(component.labels):
        x, y = label.position

        if layer_label and not (
            layer_label[0] == label.layer and layer_label[1] == label.texttype
        ):
            continue

        if get_name_from_label:
            port_name = label.text
        else:
            port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i
        if x > xc:  # east
            orientation = 0
        elif x < xc:  # west
            orientation = 180
        elif y > yc:  # north
            orientation = 90
        elif y < yc:  # south
            orientation = 270

        if fail_on_duplicates and port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )
        if get_name_from_label and port_name in component.ports:
            port_name_to_index[label.text] = (
                port_name_to_index[label.text] + 1
                if label.text in port_name_to_index
                else 1
            )
            port_name = f"{label.text}{port_name_to_index[label.text]}"

        component.add_port(
            name=port_name,
            midpoint=(x, y),
            width=port_width,
            orientation=orientation,
            port_type=port_type,
            layer=port_layer,
        )
    return component


if __name__ == "__main__":
    pass



from typing import Dict, List, Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.cross_section import strip
from gdsfactory.port import Port
from gdsfactory.routing.manhattan import round_corners
from gdsfactory.types import ComponentOrFactory, CrossSectionSpec, Route


def get_route_from_steps(
    port1: Port,
    port2: Port,
    steps: Optional[List[Dict[str, float]]] = None,
    bend: ComponentOrFactory = bend_euler,
    straight: ComponentOrFactory = straight_function,
    taper: Optional[ComponentOrFactory] = taper_function,
    cross_section: CrossSectionSpec = strip,
    **kwargs
) -> Route:
    """Returns a route formed by the given waypoints steps
    bends instead of corners and optionally tapers in straight sections.
    Tapering to wider straights reduces the optical loss.
    `get_route_from_steps` is a manual version of `get_route`
    and a more concise and convenient version of `get_route_from_waypoints`

    Args:
        port1: start port
        port2: end port
        steps: changes that define the route [{'dx': 5}, {'dy': 10}]
        bend: function that returns bends
        straight: function that returns straight waveguides
        taper: function that returns tapers
        cross_section
        **kwargs: cross_section settings

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_route_from_steps_sample")
        w = gf.components.straight()
        left = c << w
        right = c << w
        right.move((100, 80))

        obstacle = gf.components.rectangle(size=(100, 10), port_type=None)
        obstacle1 = c << obstacle
        obstacle2 = c << obstacle
        obstacle1.ymin = 40
        obstacle2.xmin = 25

        p1 = left.ports['o2']
        p2 = right.ports['o2']
        route = gf.routing.get_route_from_steps(
            port1=p1,
            port2=p2,
            steps=[
                {"x": 20},
                {"y": 20},
                {"x": 120},
                {"y": 80},
            ],
        )
        c.add(route.references)
        c.plot()
        c.show()

    """
    x, y = port1.midpoint
    x2, y2 = port2.midpoint

    waypoints = [(x, y)]
    steps = steps or []

    for d in steps:
        x = d["x"] if "x" in d else x
        x += d.get("dx", 0)
        y = d["y"] if "y" in d else y
        y += d.get("dy", 0)
        waypoints += [(x, y)]

    waypoints += [(x2, y2)]

    x = gf.get_cross_section(cross_section, **kwargs)
    auto_widen = x.auto_widen
    width1 = x.width
    width2 = x.width_wide if auto_widen else width1
    taper_length = x.taper_length
    waypoints = np.array(waypoints)

    if auto_widen:
        taper = (
            taper(
                length=taper_length,
                width1=width1,
                width2=width2,
                cross_section=cross_section,
                **kwargs,
            )
            if callable(taper)
            else taper
        )
    else:
        taper = None

    return round_corners(
        points=waypoints,
        bend=bend,
        straight=straight,
        taper=taper,
        cross_section=cross_section,
        **kwargs,
    )


@gf.cell
def test_route_from_steps() -> gf.Component:
    c = gf.Component("get_route_from_steps_sample")
    w = gf.components.straight()
    left = c << w
    right = c << w
    right.move((100, 80))

    obstacle = gf.components.rectangle(size=(100, 10))
    obstacle1 = c << obstacle
    obstacle2 = c << obstacle
    obstacle1.ymin = 40
    obstacle2.xmin = 25

    p1 = left.ports["o2"]
    p2 = right.ports["o2"]

    route = get_route_from_steps(
        port1=p1,
        port2=p2,
        steps=[
            {"x": 20, "y": 0},
            {"x": 20, "y": 20},
            {"x": 120, "y": 20},
            {"x": 120, "y": 80},
        ],
    )

    length = 186.548

    assert route.length == length, route.length
    route = gf.routing.get_route_from_steps(
        port1=p1,
        port2=p2,
        steps=[
            {"x": 20},
            {"y": 20},
            {"x": 120},
            {"y": 80},
        ],
        layer=(2, 0),
    )
    c.add(route.references)
    assert route.length == length, route.length
    return c


if __name__ == "__main__":
    # c = test_route_from_steps()
    c = gf.Component("get_route_from_steps_sample")
    w = gf.components.straight()
    left = c << w
    right = c << w
    right.move((100, 80))

    p1 = left.ports["o2"]
    p2 = right.ports["o2"]

    route = get_route_from_steps(
        port1=p2,
        port2=p1,
        # steps=[
        #     {"x": 20, "y": 0},
        #     {"x": 20, "y": 20},
        #     {"x": 120, "y": 20},
        #     {"x": 120, "y": 80},
        # ],
    )
    c.add(route.references)
    c.show()



"""High speed GSG pads."""

import gdsfactory as gf
from gdsfactory.components.pad import pad as pad_function
from gdsfactory.components.rectangle import rectangle
from gdsfactory.types import ComponentFactory, Float2, Layer

rectangle_m3 = gf.partial(rectangle, layer=gf.LAYER.M3)


@gf.cell
def pad_gsg_short(
    via_stack: ComponentFactory = rectangle_m3,
    size: Float2 = (22, 7),
    layer_metal: Layer = gf.LAYER.M3,
    metal_spacing: float = 5.0,
    short: bool = True,
    pad: ComponentFactory = pad_function,
    pad_spacing: float = 150,
) -> gf.Component:
    """Returns high speed GSG pads for calibrating the RF probes.

    Args:
        via_stack: where the RF pads connect to
        size: for the via_stack
        layer_metal:
        metal_spacing:
        short: if False returns an open
        pad: function for pad
        pad_spacing: in um
    """
    c = gf.Component()
    sig = gf.components.rectangle(size=size, layer=layer_metal).ref()

    gnd_top = c << via_stack(size=size)
    sig = via_stack(size=size)
    if short:
        sig = c << sig
    gnd_bot = c << via_stack(size=size)

    gnd_bot.ymax = sig.ymin
    gnd_top.ymin = sig.ymax

    gnd_top.movex(-metal_spacing)
    gnd_bot.movex(-metal_spacing)

    pads = c << gf.components.array(pad, columns=1, rows=3, spacing=(0, pad_spacing))
    pads.xmin = sig.xmax + 50
    pads.y = 0

    c << gf.routing.route_quad(
        gnd_bot.ports["e4"], pads.ports["e1_1_1"], layer=layer_metal
    )
    c << gf.routing.route_quad(
        gnd_top.ports["e2"], pads.ports["e1_3_1"], layer=layer_metal
    )
    c << gf.routing.route_quad(sig.ports["e3"], pads.ports["e1_2_1"], layer=layer_metal)
    return c


pad_gsg_open = gf.partial(pad_gsg_short, short=False)


if __name__ == "__main__":
    # c = pad_array_double()
    # c = pad_gsg_short()
    c = pad_gsg_open()
    c.show()



from typing import Optional

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.via_stack import via_stack_heater_m3
from gdsfactory.cross_section import strip_heater_metal, strip_heater_metal_undercut
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@cell
def straight_heater_metal_undercut(
    length: float = 320.0,
    length_undercut_spacing: float = 6.0,
    length_undercut: float = 30.0,
    length_straight_input: float = 15.0,
    heater_width: float = 2.5,
    cross_section_heater: CrossSectionSpec = strip_heater_metal,
    cross_section_heater_undercut: CrossSectionSpec = strip_heater_metal_undercut,
    with_undercut: bool = True,
    via_stack: Optional[ComponentSpec] = via_stack_heater_m3,
    port_orientation1: int = 180,
    port_orientation2: int = 0,
    heater_taper_length: Optional[float] = 5.0,
    ohms_per_square: Optional[float] = None,
    cross_section: Optional[CrossSectionSpec] = None,
    **kwargs,
) -> Component:
    """Returns a thermal phase shifter.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide
        length_undercut_spacing: from undercut regions
        length_undercut: length of each undercut section
        length_straight_input: from input port to where trenches start
        cross_section_heater: for heated sections
        cross_section_heater_undercut: for heated sections with undercut
        with_undercut: isolation trenches for higher efficiency
        via_stack: via stack
        port_orientation1: left via stack port orientation
        port_orientation2: right via stack port orientation
        heater_taper_length: minimizes current concentrations from heater to via_stack
        ohms_per_square: to calculate resistance.
        kwargs: cross_section common settings
    """
    period = length_undercut + length_undercut_spacing
    n = int((length - 2 * length_straight_input) // period)

    length_straight_input = (length - n * period) / 2

    s_si = gf.components.straight(
        cross_section=cross_section_heater,
        length=length_straight_input,
        heater_width=heater_width,
        **kwargs,
    )
    cross_section_undercut = (
        cross_section_heater_undercut if with_undercut else cross_section_heater
    )
    s_uc = gf.components.straight(
        cross_section=cross_section_undercut,
        length=length_undercut,
        heater_width=heater_width,
        **kwargs,
    )
    s_spacing = gf.components.straight(
        cross_section=cross_section_heater,
        length=length_undercut_spacing,
        heater_width=heater_width,
        **kwargs,
    )
    symbol_to_component = {
        "-": (s_si, "o1", "o2"),
        "U": (s_uc, "o1", "o2"),
        "H": (s_spacing, "o1", "o2"),
    }

    # Each character in the sequence represents a component
    sequence = "-" + n * "UH" + "-"

    c = Component()
    sequence = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component
    )
    c.add_ref(sequence)
    c.add_ports(sequence.ports)

    if via_stack:
        via_stackw = via_stacke = gf.get_component(via_stack)
        via_stack_west_midpoint = sequence.aliases["-1"].size_info.cw
        via_stack_east_midpoint = sequence.aliases["-2"].size_info.ce
        dx = via_stackw.get_ports_xsize() / 2 + heater_taper_length or 0

        via_stack_west = c << via_stackw
        via_stack_east = c << via_stacke
        via_stack_west.move(via_stack_west_midpoint - (dx, 0))
        via_stack_east.move(via_stack_east_midpoint + (dx, 0))
        c.add_port(
            "e1", port=via_stack_west.get_ports_list(orientation=port_orientation1)[0]
        )
        c.add_port(
            "e2", port=via_stack_east.get_ports_list(orientation=port_orientation2)[0]
        )
        if heater_taper_length:
            x = cross_section_heater()
            taper = gf.components.taper(
                width1=via_stackw.ports["e1"].width,
                width2=heater_width,
                length=heater_taper_length,
                layer=x.info["layer_heater"],
            )
            taper1 = c << taper
            taper2 = c << taper
            taper1.connect("o1", via_stack_west.ports["e3"])
            taper2.connect("o1", via_stack_east.ports["e1"])

    c.info["resistance"] = (
        ohms_per_square * heater_width * length if ohms_per_square else None
    )
    return c


straight_heater_metal = gf.partial(
    straight_heater_metal_undercut,
    with_undercut=False,
)
straight_heater_metal_90_90 = gf.partial(
    straight_heater_metal_undercut,
    with_undercut=False,
    port_orientation1=90,
    port_orientation2=90,
)
straight_heater_metal_undercut_90_90 = gf.partial(
    straight_heater_metal_undercut,
    with_undercut=False,
    port_orientation1=90,
    port_orientation2=90,
)


def test_ports():
    c = straight_heater_metal(length=50.0)
    assert c.ports["o2"].midpoint[0] == 50.0, c.ports["o2"].midpoint[0]
    return c


if __name__ == "__main__":
    # c = test_ports()
    # c = straight_heater_metal_undercut()
    # print(c.ports['o2'].midpoint[0])
    # c.pprint_ports()
    c = straight_heater_metal(heater_width=5, length=50.0)
    c.show()
    # scene = gf.to_trimesh(c, layer_set=gf.layers.LAYER_SET)
    # scene.show()



from gdsfactory.config import sparameters_path
from gdsfactory.simulation.simphony.model_from_sparameters import model_from_filepath


def gc1550te(filepath=sparameters_path / "gc2dte" / "gc1550.dat", numports=2):
    """Returns Sparameter model for 1550nm TE grating_coupler.

    .. plot::
        :include-source:

        import gdsfactory.simulation.simphony as gs
        import gdsfactory.simulation.simphony.components as gc

        c = gc.gc1550te()
        gs.plot_model(c)
    """
    return model_from_filepath(filepath=filepath, numports=numports)


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    import numpy as np

    wav = np.linspace(1520, 1570, 1024) * 1e-9
    f = 3e8 / wav
    c = gc1550te()
    s = c.s_parameters(freq=f)

    plt.plot(wav, np.abs(s[:, 1] ** 2))
    print(c.pins)
    plt.legend()
    plt.show()



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.bend_s import bend_s
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.types import ComponentFactory


@gf.cell
def delay_snake_sbend(
    length: float = 100.0,
    length1: float = 0.0,
    length4: float = 0.0,
    radius: float = 5.0,
    waveguide_spacing: float = 5.0,
    bend: ComponentFactory = bend_euler,
    sbend: ComponentFactory = bend_s,
    sbend_xsize: float = 100.0,
    straight: ComponentFactory = straight_function,
    **kwargs,
) -> Component:
    r"""Return compact Snake with sbend in the middle.
    Input port faces west and output port faces east.

    Args:
        length: total length
        length1: first straight section length
        length3: third straight section length
        radius: u bend radius
        waveguide_spacing: waveguide pitch
        bend:
        sbend:
        sbend_size:
        straight:
        kwargs: cross_section settings

    .. code::

                         length1
         <----------------------------
               length2    spacing    |
                _______              |
               |        \            |
               |          \          | bend1 radius
               |            \sbend   |
          bend2|              \      |
               |                \    |
               |                  \__|
               |
               ---------------------->----------->
                   length3              length4

        We adjust length2 and length3
    """

    c = Component()

    bend180_radius = (radius + waveguide_spacing) / 2
    bend = bend(radius=bend180_radius, angle=180, **kwargs)
    sbend = sbend(size=(sbend_xsize, radius), **kwargs)

    b1 = c << bend
    b2 = c << bend
    bs = c << sbend
    bs.mirror()

    length23 = (
        length - (2 * bend.info["length"] - bs.info["length"]) - length1 - length4
    )
    length2 = length23 / 2
    length3 = length23 / 2

    if length2 < 0:
        raise ValueError(
            f"length2 = {length2} < 0. You need to reduce length1 = {length1} "
            f"or length3 = {length3} or increase length = {length}"
        )

    straight1 = straight(length=length1, **kwargs)
    straight2 = straight(length=length2, **kwargs)
    straight3 = straight(length=length3, **kwargs)
    straight4 = straight(length=length4, **kwargs)

    s1 = c << straight1
    s2 = c << straight2
    s3 = c << straight3
    s4 = c << straight4

    b1.connect("o2", s1.ports["o2"])
    bs.connect("o2", b1.ports["o1"])

    s2.connect("o2", bs.ports["o1"])

    b2.connect("o1", s2.ports["o1"])
    s3.connect("o1", b2.ports["o2"])
    s4.connect("o1", s3.ports["o2"])

    c.add_port("o1", port=s1.ports["o1"])
    c.add_port("o2", port=s4.ports["o2"])

    c.info["min_bend_radius"] = sbend.info["min_bend_radius"]
    c.info["bend180_radius"] = bend180_radius
    return c


if __name__ == "__main__":
    c = gf.grid(
        [
            delay_snake_sbend(length=length, print_min_bend_radius=True)
            for length in [500, 3000]
        ]
    )
    c.show()



import numpy as np
from scipy.constants import speed_of_light
from simphony.elements import Model
from simphony.tools import interpolate

import gdsfactory as gf
import gdsfactory.simulation.lumerical as sim
from gdsfactory.component import Component


def model_from_gdsfactory(
    component: Component, dirpath=gf.CONFIG["sparameters"], **kwargs
) -> Model:
    """Return simphony model from gdsfactory Component Sparameters

    Args:
        component: component factory or instance
        dirpath: sparameters directory
        kwargs: settings
    """
    kwargs.pop("function_name", "")
    kwargs.pop("module", "")
    component = gf.call_if_func(component, **kwargs)
    pins, f, s = sim.read_sparameters_lumerical(component=component, dirpath=dirpath)

    def interpolate_sp(freq):
        return interpolate(freq, f, s)

    m = Model()
    m.pins = pins
    m.s_params = (f, s)
    m.s_parameters = interpolate_sp
    m.freq_range = (m.s_params[0][0], m.s_params[0][-1])
    m.wavelengths = speed_of_light / np.array(f)
    m.s = s
    return m


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    c = model_from_gdsfactory(gf.c.mmi2x2())
    # c = model_from_gdsfactory(gf.c.mmi2x2())
    # c = model_from_gdsfactory(gf.c.bend_euler())
    # wav = np.linspace(1520, 1570, 1024) * 1e-9
    # f = speed_of_light / wav
    # s = c.s_parameters(freq=f)

    wav = c.wavelengths
    s = c.s
    plt.plot(wav * 1e9, np.abs(s[:, 1] ** 2))
    plt.show()



import gdsfactory as gf
from gdsfactory.add_ports import add_ports_from_markers_inside
from gdsfactory.read.import_gds import import_gds


def test_import_gds_snap_to_grid() -> None:
    gdspath = gf.CONFIG["gdsdir"] / "mmi1x2.gds"
    c = import_gds(gdspath, snap_to_grid_nm=5)
    assert len(c.get_polygons()) == 8, len(c.get_polygons())

    for x, y in c.get_polygons()[0]:
        assert gf.snap.is_on_grid(x, 5)
        assert gf.snap.is_on_grid(y, 5)


def test_import_gds_hierarchy() -> gf.Component:
    c0 = gf.components.mzi_arms()
    gdspath = c0.write_gds()

    c = import_gds(gdspath)
    assert len(c.get_dependencies()) == 3, len(c.get_dependencies())
    assert c.name == c0.name, c.name
    return c


def test_import_ports() -> gf.Component:
    """Make sure you can import the ports"""
    c0 = gf.components.mzi_arms(decorator=gf.add_pins)
    gdspath = c0.write_gds()
    c0x1 = c0.ports["o1"].x
    c0x2 = c0.ports["o2"].x

    c1 = import_gds(gdspath, decorator=add_ports_from_markers_inside)
    c1x1 = c1.ports["o1"].x
    c1x2 = c1.ports["o2"].x

    assert c0x1 == c1x1
    assert c0x2 == c1x2
    return c1


# def test_import_gds_add_padding() -> gf.Component:
#     """Make sure you can import the ports"""
#     c0 = gf.components.mzi_arms(decorator=gf.add_pins)
#     gdspath = c0.write_gds()
#     gf.clear_cache()

#     c1 = import_gds(gdspath, decorator=gf.add_padding_container, name="mzi")
#     assert c1.name == "mzi"
#     return c1


if __name__ == "__main__":
    # c = test_import_gds_hierarchy()
    c = test_import_ports()
    # c = test_import_gds_add_padding()
    c.show()
    # test_import_gds_snap_to_grid()



"""avoid duplicated cell names when importing GDS files."""

import gdsfactory as gf
from gdsfactory import geometry


def test_import_first():
    c1 = gf.Component("parent")
    c1 << gf.components.mzi_arms()
    gdspath1 = c1.write_gds("extra/mzi.gds")

    mzi1 = gf.import_gds(gdspath1, safe_cell_names=True)  # IMPORT
    c1 = gf.components.mzi_arms()  # BUILD

    c2 = gf.grid([mzi1, c1])
    gdspath2 = c2.write_gds("extra/mzi2.gds")
    geometry.check_duplicated_cells(gdspath2)


def test_build_first():
    c1 = gf.Component("parent")
    c1 << gf.components.mzi_arms()
    gdspath1 = c1.write_gds("extra/mzi.gds")

    c1 = gf.components.mzi_arms()  # BUILD
    mzi1 = gf.import_gds(gdspath1, safe_cell_names=True)  # IMPORT

    c2 = gf.grid([mzi1, c1])
    gdspath2 = c2.write_gds("extra/mzi2.gds")
    geometry.check_duplicated_cells(gdspath2)


def test_import_twice():
    c0 = gf.Component("parent")
    c0 << gf.components.mzi_arms()
    gdspath1 = c0.write_gds("extra/mzi.gds")

    c1 = gf.import_gds(gdspath1)  # IMPORT
    c2 = gf.import_gds(gdspath1)  # IMPORT

    assert len(c1.references) == len(c2.references)
    assert len(c1.polygons) == len(c2.polygons)
    assert len(c1.labels) == len(c2.labels)
    assert len(c1.hash_geometry()) == len(c2.hash_geometry())


def test_import_thrice():
    c0 = gf.Component("parent")
    c0 << gf.components.mzi_arms()
    gdspath1 = c0.write_gds("extra/mzi.gds")

    c = gf.Component()
    c << gf.import_gds(gdspath1)  # IMPORT
    c << gf.import_gds(gdspath1)  # IMPORT
    c << gf.import_gds(gdspath1)  # IMPORT

    gdspath2 = c.write_gds("extra/mzis.gds")
    geometry.check_duplicated_cells(gdspath2)


if __name__ == "__main__":
    test_import_first()
    test_build_first()
    test_import_twice()
    test_import_thrice()

    # gf.clear_cache()
    # c0 << gf.components.mzi_arms()

    # gdspath1 = c0.write_gds("extra/mmi.gds")

    # c = gf.Component("parent")
    # c0 = gf.components.mmi1x2()
    # gdspath1 = "extra/mmi.gds"
    # c1 = gf.import_gds(gdspath1)  # IMPORT
    # c2 = gf.import_gds(gdspath1)  # IMPORT

    # c << c0
    # c << c1
    # c << c2

    # gdspath2 = c.write_gds("extra/mzis.gds")
    # geometry.check_duplicated_cells(gdspath2)
    # c.show()



"""Lets for example customize the default gdsfactory PDK

Fab A is mostly defined using wide layers in GDS layer (1, 0)

The metal layer traces are 2um wide

"""

import gdsfactory as gf
from gdsfactory.add_pins import add_pins_siepic
from gdsfactory.cross_section import strip

WIDTH = 2
LAYER = (1, 0)


add_pins = gf.partial(add_pins_siepic, pin_length=0.1)

xs_strip = gf.partial(
    strip, width=WIDTH, layer=LAYER, layer_bbox=(68, 0), decorator=add_pins
)

straight = gf.partial(gf.components.straight, cross_section=xs_strip)
bend_euler = gf.partial(gf.components.bend_euler, cross_section=xs_strip)
mmi1x2 = gf.partial(
    gf.components.mmi1x2,
    cross_section=xs_strip,
    width=WIDTH,
    width_taper=WIDTH,
    width_mmi=3 * WIDTH,
    decorator=add_pins,
)
ring_single = gf.partial(gf.components.ring_single, cross_section=xs_strip)
mzi = gf.partial(gf.components.mzi, cross_section=xs_strip, splitter=mmi1x2)
gc = gf.partial(
    gf.components.grating_coupler_elliptical_te,
    layer=LAYER,
    wg_width=WIDTH,
    decorator=add_pins,
)


class GenericPdk:
    straight = straight
    bend_euler = bend_euler
    mmi1x2 = mmi1x2
    ring_single = ring_single
    mzi = mzi
    gc = gc


if __name__ == "__main__":
    f = GenericPdk()

    # c = gf.components.straight(length=20, cross_section=xs_strip)
    # c = mzi()
    # c = ring_single()
    # wg_gc = gf.routing.add_fiber_array(
    #     component=c, grating_coupler=gc, cross_section=xs_strip
    # )
    # wg_gc.show(show_ports=False)



from typing import Dict, List, Optional

import numpy as np

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.components.wire import wire_corner
from gdsfactory.port import Port
from gdsfactory.routing.get_bundle_from_waypoints import get_bundle_from_waypoints
from gdsfactory.routing.sort_ports import sort_ports as sort_ports_function
from gdsfactory.types import ComponentSpec, CrossSectionSpec, Route


def get_bundle_from_steps(
    ports1: List[Port],
    ports2: List[Port],
    steps: Optional[List[Dict[str, float]]] = None,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight_function,
    taper: Optional[ComponentSpec] = taper_function,
    cross_section: CrossSectionSpec = "strip",
    sort_ports: bool = True,
    separation: Optional[float] = None,
    **kwargs
) -> List[Route]:
    """Returns a list of routes formed by the given waypoints steps
    bends instead of corners and optionally tapers in straight sections.
    Tapering to wider straights reduces the optical loss and phase errors.
    `get_bundle_from_steps` is a manual version of `get_bundle`
    and a more convenient version of `get_bundle_from_waypoints`

    Args:
        port1: start ports (list or dict)
        port2: end ports (list or dict)
        steps: changes that define the route [{'dx': 5}, {'dy': 10}]
        bend: function that returns bends
        straight: function that returns straight waveguides
        taper: function that returns tapers
        cross_section: for routes
        sort_ports: if True sort ports
        separation: center to center, defaults to ports1 separation
        kwargs: cross_section settings

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_route_from_steps_sample")
        w = gf.components.array(
            gf.partial(gf.components.straight, layer=(2, 0)),
            rows=3,
            columns=1,
            spacing=(0, 50),
        )

        left = c << w
        right = c << w
        right.move((200, 100))
        p1 = left.get_ports_list(orientation=0)
        p2 = right.get_ports_list(orientation=180)

        routes = get_bundle_from_steps(
            p1,
            p2,
            steps=[{"x": 150}],
        )

        for route in routes:
            c.add(route.references)
        c.plot()
        c.show()

    """
    if isinstance(ports1, Port):
        ports1 = [ports1]

    if isinstance(ports2, Port):
        ports2 = [ports2]

    # convert ports dict to list
    if isinstance(ports1, dict):
        ports1 = list(ports1.values())

    if isinstance(ports2, dict):
        ports2 = list(ports2.values())

    if sort_ports:
        ports1, ports2 = sort_ports_function(ports1, ports2)

    waypoints = []
    steps = steps or []

    x, y = ports1[0].midpoint
    for d in steps:
        x = d["x"] if "x" in d else x
        x += d.get("dx", 0)
        y = d["y"] if "y" in d else y
        y += d.get("dy", 0)
        waypoints += [(x, y)]

    port2 = ports2[0]
    x2, y2 = port2.midpoint
    orientation = port2.orientation

    if int(orientation) in [0, 180]:
        waypoints += [(x, y2)]
    elif int(orientation) in [90, 270]:
        waypoints += [(x2, y)]

    x = gf.get_cross_section(cross_section, **kwargs)
    auto_widen = x.auto_widen
    width1 = x.width
    width2 = x.width_wide if auto_widen else width1
    taper_length = x.taper_length
    waypoints = np.array(waypoints)

    if auto_widen:
        taper = gf.get_component(
            taper,
            length=taper_length,
            width1=width1,
            width2=width2,
            cross_section=cross_section,
            **kwargs,
        )
    else:
        taper = None

    return get_bundle_from_waypoints(
        ports1=ports1,
        ports2=ports2,
        waypoints=waypoints,
        bend=bend,
        straight=straight,
        taper=taper,
        cross_section=cross_section,
        separation=separation,
        **kwargs,
    )


get_bundle_from_steps_electrical = gf.partial(
    get_bundle_from_steps, bend=wire_corner, cross_section=gf.cross_section.metal3
)


def _demo():
    c = gf.Component("get_route_from_steps_sample")

    w = gf.components.array(
        gf.partial(gf.components.straight, layer=(2, 0)),
        rows=3,
        columns=1,
        spacing=(0, 50),
    )

    left = c << w
    right = c << w
    right.move((200, 100))
    p1 = left.get_ports_list(orientation=0)
    p2 = right.get_ports_list(orientation=180)

    routes = get_bundle_from_steps_electrical(
        p1,
        p2,
        steps=[{"x": 150}],
    )

    for route in routes:
        c.add(route.references)

    c.show()


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("pads_bundle_steps")
    pt = c << gf.components.pad_array(
        gf.partial(gf.components.pad, size=(30, 30)),
        orientation=270,
        columns=3,
        spacing=(50, 0),
    )
    pb = c << gf.components.pad_array(orientation=90, columns=3)
    pt.move((300, 500))

    routes = gf.routing.get_bundle_from_steps_electrical(
        pb.ports, pt.ports, end_straight_length=60, separation=30, steps=[{"dy": 100}]
    )

    for route in routes:
        c.add(route.references)

    c.show()



"""You can define a function to add pins """
import gdsfactory as gf
from gdsfactory.add_pins import add_pins_triangle
from gdsfactory.component import Component


def test_pins_custom() -> Component:
    """You can define the `pins_function` that we use to add markers to each port"""
    c = gf.components.straight(length=11.1)
    c = c.copy()
    add_pins_triangle(component=c)
    return c


if __name__ == "__main__":
    c = test_pins_custom()
    c.show(show_ports=False)



from collections import OrderedDict
from typing import Optional, Tuple

import numpy as np
from omegaconf import DictConfig

from gdsfactory import Component

c = 2.9979e8
pi = np.pi
um = 1e-6


def set_global_settings(session: object, simulation_settings: dict):
    for param, val in zip(simulation_settings.keys(), simulation_settings.values()):
        session.setnamed("::Root Element", param, val)


def add_interconnect_element(
    session: object,
    label: str,
    model: str,
    loc: Tuple[int, int] = (200.0, 200.0),
    flip_vert: bool = False,
    flip_horiz: bool = False,
    rotation: float = 0.0,
    extra_props: OrderedDict = None,
):
    """
    Add an element to the Interconnect session.
    TODO: Need to connect this to generated s-parameters and add them to the model as well

    Args:
        session: Interconnect session
        label: label for Interconnect component
        model:
        loc:
        flip_vert:
        flip_horiz:
        rotation:
        extra_props:

    """
    props = OrderedDict(
        [
            ("name", label),
            ("x position", loc[0]),
            ("y position", loc[1]),
            ("horizontal flipped", float(flip_horiz)),
            ("vertical flipped", float(flip_vert)),
            ("rotated", rotation),
        ]
    )
    if extra_props:
        props.update(extra_props)
    return session.addelement(model, properties=props)


def send_to_interconnect(
    component: Component,
    session: Optional[object] = None,
    placements: dict = None,
    simulation_settings: OrderedDict = None,
    drop_port_prefix: str = None,
    component_distance_scaling: float = 1,
    **settings,
) -> object:
    """Send all components in netlist to Interconnect and make connections according to netlist.

    Args:
        component: component from which to extract netlist
        session: Interconnect session
        placements: x,y pairs for where to place the components in the Interconnect GUI
        simulation_settings: global settings for Interconnect simulation
        drop_port_prefix: if components are written with some prefix, drop up to and including
            the prefix character.  (i.e. "c1_input" -> "input")
        component_distance_scaling: scaling factor for component distances when laying out Interconnect schematic
    """
    import sys

    if "lumapi" not in sys.modules.keys():
        import lumapi

    inc = session or lumapi.INTERCONNECT(hide=False)

    inc.switchtolayout()
    inc.deleteall()

    c = component

    netlist = c.get_netlist()

    instances: DictConfig = netlist["instances"]
    connections: DictConfig = netlist["connections"]
    placements: DictConfig = netlist["placements"] if not placements else placements

    for i, instance in enumerate(instances):
        info = instances[instance].info
        extra_props = info["interconnect"] if "interconnect" in info.keys() else None
        add_interconnect_element(
            session=inc,
            label=instance,
            model=info.model,
            loc=(
                component_distance_scaling * placements[instance].x,
                component_distance_scaling * placements[instance].y,
            ),
            rotation=placements[instance].rotation,
            extra_props=extra_props,
        )

    for connection in connections:
        element2, port2 = connection.split(",")
        element1, port1 = connections[connection].split(",")

        if drop_port_prefix:
            # a bad way to autodetect which ports need to have prefixes dropped..
            try:
                port1 = port1[port1.index(drop_port_prefix) + 1 :]
            except:
                pass
            try:
                port2 = port2[port2.index(drop_port_prefix) + 1 :]
            except:
                pass

        # EBeam ports are not named consistently between Klayout and Interconnect..
        if hasattr(instances[element1].info, port1):
            port1 = instances[element1].info[port1]
        if hasattr(instances[element2].info, port2):
            port2 = instances[element2].info[port2]

        inc.connect(element1, port1, element2, port2)

    if simulation_settings:
        set_global_settings(inc, simulation_settings)
    return inc


def run_wavelength_sweep(
    component: Component,
    session: Optional[object] = None,
    port_in: Tuple = None,
    ports_out: Tuple = None,
    wavelength_range: Tuple = (1.500, 1.600),
    n_points: int = 1000,
    results: Tuple = ("transmission",),
    extra_ona_props: dict = None,
    **kwargs,
) -> dict:
    """
    Args:
        component:
        session:
        port_in: specify the port in the Interconnect model to attach the ONA output to
        ports_out: specify the ports in the Interconnect models to attach the ONA input to
        wavelength_range:
        n_points:
        results:
        extra_ona_props:
        kwargs:

    """
    import lumapi

    inc = session if session else lumapi.INTERCONNECT()

    # Add Monte-Carlo params
    inc.addproperty("::Root Element", "MC_uniformity_thickness", "wafer", "Matrix")
    inc.addproperty("::Root Element", "MC_uniformity_width", "wafer", "Matrix")
    inc.addproperty("::Root Element", "MC_non_uniform", "wafer", "Number")
    inc.addproperty("::Root Element", "MC_grid", "wafer", "Number")
    inc.addproperty("::Root Element", "MC_resolution_x", "wafer", "Number")
    inc.addproperty("::Root Element", "MC_resolution_y", "wafer", "Number")

    # send circuit to interconnect
    inc = send_to_interconnect(component=component, session=inc, **kwargs)

    ona_props = OrderedDict(
        [
            ("number of input ports", len(ports_out)),
            ("number of points", n_points),
            ("input parameter", "start and stop"),
            ("start frequency", (c / (wavelength_range[1] * um))),
            ("stop frequency", (c / (wavelength_range[0] * um))),
            ("plot kind", "wavelength"),
            ("relative to center", float(False)),
        ]
    )
    if extra_ona_props:
        ona_props.update(extra_ona_props)

    ona = add_interconnect_element(
        session=inc,
        model="Optical Network Analyzer",
        label="ONA_1",
        loc=(0, -50),
        extra_props=ona_props,
    )

    inc.connect(ona.name, "output", *port_in)
    for i, port in enumerate(ports_out):
        inc.connect(ona.name, f"input {i+1}", *ports_out[i])

    inc.run()

    data = dict()
    for result in results:
        data[result] = {
            port: inc.getresult(ona.name, f"input {i+1}/mode 1/{result}")
            for i, port in enumerate(ports_out)
        }

    # inc.close()
    return data


if __name__ == "__main__":
    import ubcpdk.components as pdk

    import gdsfactory as gf

    component = gf.Component()
    gc1 = component << pdk.gc_te1550()
    gc2 = component << pdk.gc_te1550()

    s = component << pdk.y_splitter()

    gc1.connect(port="opt1", destination=s.ports["opt1"])
    gc2.connect(port="opt1", destination=s.ports["opt2"])

    component.show()

    netlist = component.get_netlist()

    simulation_settings = OrderedDict(
        [
            ("MC_uniformity_thickness", np.array([200, 200])),
            ("MC_uniformity_width", np.array([200, 200])),
            ("MC_non_uniform", 0),
            ("MC_grid", 1e-5),
            ("MC_resolution_x", 200),
            ("MC_resolution_y", 0),
        ]
    )

    # inc = send_to_interconnect(
    #     component=c,
    #     simulation_settings=simulation_settings
    # )
    from gdsfactory.get_netlist import get_instance_name

    gc1_netlist_instance_name = get_instance_name(component, gc1)
    gc2_netlist_instance_name = get_instance_name(component, gc2)

    ports_out = (
        (gc2_netlist_instance_name, "opt_fiber"),
        (gc1_netlist_instance_name, "opt_wg"),
    )
    results = run_wavelength_sweep(
        component=component,
        port_in=(gc1_netlist_instance_name, "opt_fiber"),
        ports_out=ports_out,
        simulation_settings=simulation_settings,
        results=("transmission",),
    )

    import matplotlib.pyplot as plt

    plt.figure()
    for port in ports_out:
        wl = results["transmission"][port]["wavelength"] / um
        T = 10 * np.log10(np.abs(results["transmission"][port]["TE transmission"]))
        plt.plot(wl, T, label=f"{port}")
    plt.legend()
    plt.xlabel(r"Wavelength ($\mu$m)")
    plt.ylabel("TE transmission (dB)")
    plt.show()
    pass



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.bend_s import bend_s
from gdsfactory.cross_section import strip
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@gf.cell
def coupler_symmetric(
    bend: ComponentSpec = bend_s,
    gap: float = 0.234,
    dy: float = 5.0,
    dx: float = 10.0,
    cross_section: CrossSectionSpec = strip,
    **kwargs,
) -> Component:
    r"""Two coupled straights with bends.

    Args:
        bend: bend or library
        gap:
        dy: port to port vertical spacing
        dx: bend length in x direction
        cross_section:
        **kwargs: cross_section settings

    .. code::

                       dx
                    |-----|
                       ___ E1
                      /       |
                _____/        |
           gap  _____         |  dy
                     \        |
                      \___    |
                           E0

    """

    x = gf.get_cross_section(cross_section, **kwargs)
    width = x.width
    bend_component = gf.get_component(
        bend,
        size=(dx, (dy - gap - width) / 2),
        cross_section=cross_section,
        **kwargs,
    )

    w = bend_component.ports["o1"].width
    y = (w + gap) / 2

    c = Component()
    top_bend = bend_component.ref(position=(0, y), port_id="o1")
    bottom_bend = bend_component.ref(position=(0, -y), port_id="o1", v_mirror=True)

    c.add(top_bend)
    c.add(bottom_bend)

    c.absorb(top_bend)
    c.absorb(bottom_bend)

    c.add_port("o1", port=bottom_bend.ports["o1"])
    c.add_port("o2", port=top_bend.ports["o1"])

    c.add_port("o3", port=top_bend.ports["o2"])
    c.add_port("o4", port=bottom_bend.ports["o2"])
    c.info["length"] = bend_component.info["length"]
    c.info["min_bend_radius"] = bend_component.info["min_bend_radius"]
    return c


if __name__ == "__main__":
    c = coupler_symmetric(gap=0.2, width=0.9)
    c.show()
    c.pprint()

    for dyi in [2, 3, 4, 5]:
        c = coupler_symmetric(gap=0.2, width=0.5, dy=dyi, dx=10.0, layer=(2, 0))
        print(f"dy={dyi}, min_bend_radius = {c.info['min_bend_radius']}")



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component


def test_path_length_matching() -> Component:
    c = gf.Component("path_length_match_sample")

    dy = 2000.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 100.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N)]

    a1 = 90
    a2 = a1 + 180

    layer = (1, 0)
    ports1 = [
        gf.Port(
            f"top_{i}", midpoint=(xs1[i], 0), width=0.5, orientation=a1, layer=layer
        )
        for i in range(N)
    ]
    ports2 = [
        gf.Port(
            f"bot_{i}", midpoint=(xs2[i], dy), width=0.5, orientation=a2, layer=layer
        )
        for i in range(N)
    ]

    routes = gf.routing.get_bundle_path_length_match(ports1, ports2)
    lengths = [2659.822]
    for route, length in zip(routes, lengths):
        c.add(route.references)
        assert np.isclose(route.length, length), route.length
    return c


def test_path_length_matching_extra_length() -> Component:
    c = gf.Component("path_length_match_sample")

    dy = 2000.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 100.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N)]

    a1 = 90
    a2 = a1 + 180
    layer = (1, 0)

    ports1 = [
        gf.Port(
            f"top_{i}", midpoint=(xs1[i], 0), width=0.5, orientation=a1, layer=layer
        )
        for i in range(N)
    ]
    ports2 = [
        gf.Port(
            f"bot_{i}", midpoint=(xs2[i], dy), width=0.5, orientation=a2, layer=layer
        )
        for i in range(N)
    ]

    routes = gf.routing.get_bundle_path_length_match(ports1, ports2, extra_length=40)
    lengths = [2699.822]
    for route, length in zip(routes, lengths):
        c.add(route.references)
        assert np.isclose(route.length, length), route.length
    return c


def test_path_length_matching_nb_loops() -> Component:
    c = gf.Component("path_length_match_sample")

    dy = 2000.0
    xs1 = [-500, -300, -100, -90, -80, -55, -35, 200, 210, 240, 500, 650]

    pitch = 100.0
    N = len(xs1)
    xs2 = [-20 + i * pitch for i in range(N)]

    a1 = 90
    a2 = a1 + 180

    layer = (1, 0)
    ports1 = [
        gf.Port(
            f"top_{i}", midpoint=(xs1[i], 0), width=0.5, orientation=a1, layer=layer
        )
        for i in range(N)
    ]
    ports2 = [
        gf.Port(
            f"bot_{i}", midpoint=(xs2[i], dy), width=0.5, orientation=a2, layer=layer
        )
        for i in range(N)
    ]

    routes = gf.routing.get_bundle_path_length_match(ports1, ports2, nb_loops=2)
    lengths = [2686.37]
    for route, length in zip(routes, lengths):
        c.add(route.references)
        assert np.isclose(route.length, length), route.length
    return c


if __name__ == "__main__":
    c = test_path_length_matching()
    # c = test_path_length_matching_extra_length()
    # c = test_path_length_matching_nb_loops()
    c.show()
